﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Text;
using PaintDotNet;
using PaintDotNet.Effects;
using PaintDotNet.IndirectUI;
using PaintDotNet.PropertySystem;

namespace HQFontEffect
{
    [PluginSupportInfo(typeof(PluginSupportInfo), DisplayName = "High Quality Font")]
    public class HQFont : PropertyBasedEffect
    {
        public static readonly string EffectName = "High Quality Font";
        public static Image EffectIcon
        {
            get
            {
                Surface s = new Surface(16, 16);
                for (int j = 0; j < 16; j++)
                    for (int i = 0; i < 16; i++)
                    {
                        byte b = (byte)(255 * ((i % 2) ^ (j % 2)));
                        s[i, j] = ColorBgra.FromBgr(b, b, b);
                    }
                return s.CreateAliasedBitmap();
            }
        } // Shrug, it's short and sweet.

        public HQFont()
            : base(EffectName, EffectIcon, SubmenuNames.Render, EffectFlags.Configurable)
        {
        }

        public enum eC
        {
            text=1,
            fontName=2,
            fontSize=4
        }

        protected override PropertyCollection OnCreatePropertyCollection()
        {
            List<Property> props = new List<Property>();

            props.Add(new StringProperty(eC.text,""));
            FontFamily[] ff = new InstalledFontCollection().Families;
            props.Add(new StaticListChoiceProperty(eC.fontName,ff,0,false));
            props.Add(new Int32Property(eC.fontSize,19,1,1280));

            return new PropertyCollection(props);
        }

        protected override ControlInfo OnCreateConfigUI(PropertyCollection props)
        {
            //return base.OnCreateConfigUI(props);
            ControlInfo c = CreateDefaultConfigUI(props);
            

            PropertyControlInfo i = c.FindControlForPropertyName(eC.fontName);
            FontFamily[] ffs = new InstalledFontCollection().Families;
            foreach (FontFamily ff in ffs)
                i.SetValueDisplayName(ff, ff.Name);
            return c;
        }

        protected override void OnSetRenderInfo(PropertyBasedEffectConfigToken newToken, RenderArgs dstArgs, RenderArgs srcArgs)
        {
            this.text = newToken.GetProperty<StringProperty>(eC.text).Value;
            this.font = (FontFamily)newToken.GetProperty<StaticListChoiceProperty>(eC.fontName).Value;
            this.fontSize = newToken.GetProperty<Int32Property>(eC.fontSize).Value;
            validated = false;

            base.OnSetRenderInfo(newToken, dstArgs, srcArgs);
        }

        protected override void OnRender(Rectangle[] renderRects, int startIndex, int length)
        {
            if (length==0) return;
            Surface surf;
            ColorBgra srcColor,inColor;
            UserBlendOp normBlend = new UserBlendOps.NormalBlendOp();
            for (int i = startIndex; i < startIndex + length; ++i)
            {
                int y0 = renderRects[i].Top;
                int x0 = renderRects[i].Left;
                int y1 = renderRects[i].Bottom;
                int x1 = renderRects[i].Right;
                for (int v = y0; v < y1; v++)
                {
                    for (int u = x0; u < x1; u++)
                    {
                        if (stopRendering > v)
                        {
                            srcColor = SrcArgs.Surface[u, v];
                            inColor = srcColor;
                            surf = GetMappedField(u, v, ss);
                            if (surf != null)
                                inColor = SamplePixel(surf);
                            DstArgs.Surface[u, v] = normBlend.Apply(srcColor, inColor);
                        }
                        else
                        {
                            DstArgs.Surface[u, v] = SrcArgs.Surface[u, v];
                        }
                    }
                }
            }
        }
        /*private List<PointF> ourPointField
        {
            get
            {
                List<PointF> pF = new List<PointF>(14);
                

                pF.Add(new PointF(

            }
        }*/ // Actually...
        private Surface GetMappedField(int xPiece, int yPiece, int ss) // computes a 512x512 section at a time o:
        {
            float x, y;
            x = - (long)(xPiece * ss); // Meh!
            y = - (long)(yPiece * ss);
            Surface s = xParentSurface;
            // our surface starts at xPiece*512, yPiece*512 in a field such as
            //0,0 … 0,1 {xPiece,yPiece}
            // ⋮     ⋮
            //1,0 … 1,1
            Graphics g = new RenderArgs(s).Graphics;
            //ColorBgra PC = (ColorBgra)EnvironmentParameters.PrimaryColor;
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.None; // Gives the user more control! >.>
            SolidBrush color = new SolidBrush(Color.FromArgb(255,0,0,0));
            FontStyle fs = (font.IsStyleAvailable(FontStyle.Regular))?FontStyle.Regular:
                ((font.IsStyleAvailable(FontStyle.Italic))?FontStyle.Italic:
                ((font.IsStyleAvailable(FontStyle.Bold))?FontStyle.Bold:
                ((font.IsStyleAvailable(FontStyle.Underline))?FontStyle.Underline:
                ((font.IsStyleAvailable(FontStyle.Strikeout))?FontStyle.Strikeout:
                0))));
            if (fs == 0) return null;
            g.DrawString(text, new Font(font.Name, fontSize*ss, fs), color, x, y);
            return s;
        }
        private ColorBgra SamplePixel(Surface s)
        {
            double r = 0;
            double g = 0;
            double b = 0;
            double a = 0;
            int w = s.Width;
            int h = s.Height;
            for (int j = 0; j < h; j++)
            {
                for (int i = 0; i < w; i++)
                {
                    r += ((double)s[i, j].R)/ss2;
                    g += ((double)s[i, j].G)/ss2;
                    b += ((double)s[i, j].B)/ss2;
                    a += ((double)s[i, j].A)/ss2;
                }
            }
            byte R = (byte)r;
            byte G = (byte)g;
            byte B = (byte)b;
            byte A = (byte)a;
            return ColorBgra.FromBgra(B, G, R, A);
        }
        private string text;
        private FontFamily font;
        private int fontSize;
        public int stopRendering
        {
            get
            {
                if (_stopRendering != 0 && validated)
                    return _stopRendering;
                else
                {
                    int c = 1;
                    int startIndex = 0;
                    while (text.IndexOf("\n", startIndex) >= 0)
                    {
                        c++;
                        startIndex = text.IndexOf("\n", startIndex) + 1;
                    }
                    _stopRendering = c * 2 * fontSize;
                    return c * 2 * fontSize;
                }
            }
        }
        private int _stopRendering = 0;
        private bool validated = true;
        public Surface xParentSurface
        {
            get
            {
                if (_xPS != null)
                    return _xPS.Clone();
                else
                {
                    Surface s = new Surface(ss,ss);
                    for (int j=0; j<ss; j++)
                        for (int i=0; i<ss; i++)
                            s[i,j] = ColorBgra.Transparent;
                    _xPS = s;
                    return s;
                }
            }
        }
        private int ss = 16;
        private double ss2 = 256; //ss^2
        private Surface _xPS = null;
    }
}
